@@ -56,46 +56,6 @@ class CacheTagging(object):
             self.set(name, value, tags, timeout, version)
         return value
 
-    def set(self, name, value, tags=(), timeout=None, version=None):
-        """Sets cache value and tags."""
-        if not hasattr(tags, '__iter__'):  # Called as native API
-            if timeout is not None and version is None:
-                version = timeout
-            timeout = tags
-            return self.cache.set(name, value, timeout, version)
-
-        # pull tags from descendants (cached fragments)
-        tags = set(tags)
-        for tag, vers in self.get_ancestors().get(name, set()):
-            if vers == version:
-                tags.add(tag)
-
-        tag_versions = {}
-        if len(tags):
-            tag_caches = self.cache.get_many(
-                list(map(tag_prepare_name, tags))
-            )
-            tag_new_dict = {}
-            for tag in tags:
-                tag_prepared = tag_prepare_name(tag)
-                if tag_prepared not in tag_caches\
-                        or tag_caches[tag_prepared] is None:
-                    tag_version = tag_generate_version()
-                    tag_new_dict[tag_prepared] = tag_version
-                else:
-                    tag_version = tag_caches[tag_prepared]
-                tag_versions[tag] = tag_version
-            if len(tag_new_dict):
-                self.cache.set_many(tag_new_dict, TAG_TIMEOUT, version)
-
-        data = {
-            'tag_versions': tag_versions,
-            'value': value,
-        }
-
-        self.finish(name, tags, version=version)
-        return self.cache.set(name, data, timeout, version)
-
     def get(self, name, default=None, version=None, abort=False):
         """Gets cache value.
 
@@ -125,12 +85,52 @@ class CacheTagging(object):
         self.finish(name, data['tag_versions'].keys(), version=version)
         return data['value']
 
+    def set(self, name, value, tags=(), timeout=None, version=None):
+        """Sets cache value and tags."""
+        if not hasattr(tags, '__iter__'):  # Called as native API
+            if timeout is not None and version is None:
+                version = timeout
+            timeout = tags
+            return self.cache.set(name, value, timeout, version)
+
+        tags = set(tags)
+        # pull tags from descendants (cached fragments)
+        try:
+            tags.update(self.get_ancestors()[name][version])
+        except KeyError:
+            pass
+
+        tag_versions = {}
+        if len(tags):
+            tag_caches = self.cache.get_many(
+                list(map(tag_prepare_name, tags))
+            )
+            tag_new_dict = {}
+            for tag in tags:
+                tag_prepared = tag_prepare_name(tag)
+                if tag_prepared not in tag_caches\
+                        or tag_caches[tag_prepared] is None:
+                    tag_version = tag_generate_version()
+                    tag_new_dict[tag_prepared] = tag_version
+                else:
+                    tag_version = tag_caches[tag_prepared]
+                tag_versions[tag] = tag_version
+            if len(tag_new_dict):
+                self.cache.set_many(tag_new_dict, TAG_TIMEOUT, version)
+
+        data = {
+            'tag_versions': tag_versions,
+            'value': value,
+        }
+
+        self.finish(name, tags, version=version)
+        return self.cache.set(name, data, timeout, version)
+
     def invalidate_tags(self, *tags, **kwargs):
         """Invalidate specified tags"""
         if len(tags):
             version = kwargs.get('version', None)
-            tags = set(tags)
-            tags_prepared = list(map(tag_prepare_name, tags))
+            tags_prepared = list(map(tag_prepare_name, set(tags)))
             self.add_to_transaction_scope(tags_prepared, version=version)
             self.cache.delete_many(tags_prepared, version=version)
 
@@ -142,26 +142,25 @@ class CacheTagging(object):
 
     def add_to_ancestors(self, tags, version=None):
         """add tags to ancestor"""
-        for k, v in self.get_ancestors().items():
-            for tag in tags:
-                v.add((tag, version,))
+        for cachename, versions in self.get_ancestors().items():
+            versions.setdefault(version, set()).update(tags)
 
     def begin(self, name):
         """Start cache creating."""
-        self.get_ancestors()[name] = set()
+        self.get_ancestors()[name] = {}
 
     def abort(self, name):
-        """Clean tags for given template name."""
-        self.get_ancestors().pop(name, set())
+        """Clean tags for given cache name."""
+        self.get_ancestors().pop(name, {})
 
     def finish(self, name, tags, version=None):
         """Start cache creating."""
-        self.get_ancestors().pop(name, set())
+        self.get_ancestors().pop(name, {})
         self.add_to_ancestors(tags, version=version)
 
     def transaction_begin(self):
         """Handles database transaction begin."""
-        self.get_transaction_scopes().append([])
+        self.get_transaction_scopes().append({})
         return self
 
     def transaction_finish(self):
@@ -175,12 +174,8 @@ class CacheTagging(object):
         or "transaction_rollback")."""
         scope = self.get_transaction_scopes().pop()
         if len(scope):
-            scope_versioned = {}
-            for tag, version in scope:
-                tags = scope_versioned.setdefault(version, [])
-                tags.append(tag)
-            for version, tags in scope_versioned.items():
-                self.cache.delete_many(list(set(tags)), version=version)
+            for version, tags in scope.items():
+                self.cache.delete_many(list(tags), version=version)
         return self
 
     def transaction_finish_all(self):
@@ -199,9 +194,7 @@ class CacheTagging(object):
         """Adds cache names to current scope."""
         scopes = self.get_transaction_scopes()
         if len(scopes):
-            scope = scopes[-1]
-            for tag in tags:
-                scope.append((tag, version,))
+            scopes[-1].setdefault(version, set()).update(tags)
 
     def __getattr__(self, name):
         """Proxy for all native methods."""
